/*
 * Copyright (C) 2005 Luca Veltri - University of Parma - Italy
 * 
 * SipToSis is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * SipToSis is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this source code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Author(s):
 * Luca Veltri (luca.veltri@unipr.it)
 * 
 * Greg Dorfuss - changed long and int functions
 *              - copied here so some things can be directly accessed
 *              
 * Greg Dorfuss 2009/02/07 - fix setbit and getbit functions
 *                         - added setDuration/getDuration functions for rfc2833 use    
 *              2009/12/22 - removed checking for >=12 packet length that can't occur without some really stupid code 
 */

package local.ua;

import org.zoolu.tools.Random;

/** RtpPacket implements a RTP packet. 
 */
public class RtpPacket
{
   /* RTP packet buffer containing both the RTP header and payload */   
   byte[] packet;
   
   /* RTP packet length */   
   int packet_len;

   /* RTP header length */   
   //int header_len;
   
   private static boolean bigEndian=true;

   /** Gets the RTP packet */
   final public byte[] getPacket()
   {  return packet;
   }

   /** Gets the RTP packet length */
   final public int getLength()
   {  return packet_len;
   }

   /** Gets the RTP header length */
   final public int getHeaderLength()
   {  
	  return 12+4*(packet[0] & 0x0F);
   }

   /** Gets the RTP header length */
   final public int getPayloadLength()
   {  
	   return packet_len-getHeaderLength();
   }

   /** Sets the RTP payload length */
   final public void setPayloadLength(int len)
   {  packet_len=getHeaderLength()+len;
   }

   // version (V): 2 bits
   // padding (P): 1 bit
   // extension (X): 1 bit
   // CSRC count (CC): 4 bits
   // marker (M): 1 bit
   // payload type (PT): 7 bits
   // sequence number: 16 bits
   // timestamp: 32 bits
   // SSRC: 32 bits
   // CSRC list: 0 to 15 items, 32 bits each


   /** Gets the version (V) */
   final public int getVersion()
   {  
	   return (packet[0]>>6 & 0x03);
   }

   /** Sets the version (V) */
   final public void setVersion(int v)
   {  
	   packet[0]=(byte)((packet[0] & 0x3F) | ((v & 0x03)<<6));
   }

   /** Whether has padding (P) */
   final public boolean hasPadding()
   {  
	  return getBit(packet[0],5);
   }

   /** Set padding (P) */
   final public void setPadding(boolean p)
   {  packet[0]=setBit(p,packet[0],5);
   }

   /** Whether has extension (X) */
   final public boolean hasExtension()
   {  return getBit(packet[0],4);
   }

   /** Set extension (X) */
   final public void setExtension(boolean x)
   {  packet[0]=setBit(x,packet[0],4);
   }

   /** Gets the CSCR count (CC) */
   final public int getCscrCount()
   {  return (packet[0] & 0x0F);
   }

   /** Whether has marker (M) */
   final public boolean hasMarker()
   {  return getBit(packet[1],7);
   }

   /** Set marker (M) */
   final public void setMarker(boolean m)
   { 
	   packet[1]=setBit(m,packet[1],7);
   }

   /** Gets the payload type (PT) */
   final public int getPayloadType()
   {  return (packet[1] & 0x7F);
   }

   /** Sets the payload type (PT) */
   final public void setPayloadType(int pt)
   {  packet[1]=(byte)((packet[1] & 0x80) | (pt & 0x7F));
   }

   /** Gets the sequence number */
   final public int getSequenceNumber()
   {  return getInt(packet,2);
   }

   /** Sets the sequence number */
   final public void setSequenceNumber(int sn)
   {  setInt(sn,packet,2);
   }

   /** Gets the timestamp */
   final public long getTimestamp()
   {  return getLong(packet,4);
   }

   /** Sets the timestamp */
   final public void setTimestamp(long timestamp)
   {  setLong(timestamp,packet,4);
   }
   
   /** Gets the SSCR */
   final public long getSscr()
   {  return getLong(packet,8);
   }

   /** Sets the SSCR */
   final public void setSscr(long ssrc)
   {  setLong(ssrc,packet,8);
   }

   /** Gets the CSCR list */
   final public long[] getCscrList()
   {  int cc=getCscrCount();
      long[] cscr=new long[cc];
      for (int i=0; i<cc; i++) cscr[i]=getLong(packet,12+4*i); 
      return cscr;
   }

   /** Sets the CSCR list */
   final public void setCscrList(long[] cscr)
   {    int cc=cscr.length;
         if (cc>15) cc=15;
         packet[0]=(byte)(((packet[0]>>4)<<4)+cc);
         cscr=new long[cc];
         for (int i=0; i<cc; i++) setLong(cscr[i],packet,12+4*i);   
         //header_len=12+4*cc;
   }

   /** Sets the payload */
   final public void setPayload(byte[] payload, int len)
   {  
	     int header_len=getHeaderLength();
         //for (int i=0; i<len; i++) packet[header_len+i]=payload[i]; 
         System.arraycopy(payload, 0, packet, header_len,  len);
         packet_len=header_len+len;
         
   }

   /** Gets the payload */
   final public byte[] getPayload()
   {  int header_len=getHeaderLength();
      int len=packet_len-header_len;
      byte[] payload=new byte[len];
      //for (int i=0; i<len; i++) payload[i]=packet[header_len+i]; 
      System.arraycopy(packet, header_len, payload, 0,len);
      return payload;
   }

   /** Creates a new RTP packet */ 
   public RtpPacket(byte[] buffer, int packet_length)
   {  
      packet=buffer;
      packet_len=packet_length;
      if (packet_len<12) packet_len=12;
      init(0x0F);
   }


   /** init the RTP packet header (only PT) */ 
   final public void init(int ptype)
   {  init(ptype,Random.nextLong());
   }


   /** init the RTP packet header (PT and SSCR) */ 
   final public void init(int ptype, long sscr)
   {  init(ptype,Random.nextInt(),Random.nextLong(),sscr);
   }
   

   /** init the RTP packet header (PT, SQN, TimeStamp, SSCR) */ 
   final public void init(int ptype, int seqn, long timestamp, long sscr)
   {  setVersion(2);
      setPayloadType(ptype);
      setSequenceNumber(seqn);
      setTimestamp(timestamp);
      setSscr(sscr);
   }

   // *********************** Private and Static ***********************

   /* used to set duration bits for rfc2833 packets */
   final public void setDuration(int dur)
   {
	   setInt(dur,packet,getHeaderLength()+2);
   }
   final public int getDuration()
   {
	   return getInt(packet,getHeaderLength()+2);
   }

   private static int getLong(byte buffer[], int byteOffset)
   {
       return bigEndian ? buffer[byteOffset] << 24 | (buffer[byteOffset + 1] & 0xff) << 16 | (buffer[byteOffset + 2] & 0xff) << 8 | buffer[byteOffset + 3] & 0xff : buffer[byteOffset + 3] << 24 | (buffer[byteOffset + 2] & 0xff) << 16 | (buffer[byteOffset + 1] & 0xff) << 8 | buffer[byteOffset] & 0xff;
   }
   
   
   private static void setLong(long sample, byte buffer[], int byteOffset)
   {
       if(bigEndian)
       {
           buffer[byteOffset++] = (byte)(sample >> 24);
           buffer[byteOffset++] = (byte)(sample >>> 16 & 0xff);
           buffer[byteOffset++] = (byte)(sample >>> 8 & 0xff);
           buffer[byteOffset] = (byte)(sample & 0xff);
       } else
       {
           buffer[byteOffset++] = (byte)(sample & 0xff);
           buffer[byteOffset++] = (byte)(sample >>> 8 & 0xff);
           buffer[byteOffset++] = (byte)(sample >>> 16 & 0xff);
           buffer[byteOffset] = (byte)(sample >> 24);
       }
   }

 
   private static int getInt(byte buffer[], int byteOffset)
   {
       return bigEndian ? buffer[byteOffset] << 8 | buffer[byteOffset + 1] & 0xff : buffer[byteOffset + 1] << 8 | buffer[byteOffset] & 0xff;
   }

   private static void setInt(int sample, byte buffer[], int byteOffset)
   {
       if(bigEndian)
       {
           buffer[byteOffset++] = (byte)(sample >> 8);
           buffer[byteOffset] = (byte)(sample & 0xff);
       } else
       {
           buffer[byteOffset++] = (byte)(sample & 0xff);
           buffer[byteOffset] = (byte)(sample >> 8);
       }
   }
   
   
    /** Gets bit value */
   private static boolean getBit(byte b, int bit)
   {  
	   return !((b>>bit)==0);
   }

   /** Sets bit value */
   private static byte setBit(boolean value, byte b, int bit)
   {  
	  if 
	     (value) b=(byte)(b|(1<<bit));
      else 
    	  b=(byte)((b|(1<<bit))^(1<<bit));

      return b;
   }


   /** testing 
	public static void main(String[] args) throws Exception 
   {
		byte[] dtmfbuf=new byte[512];
		RtpPacket dtmfpacket=new RtpPacket(dtmfbuf,0);
		
		System.out.println(dtmfbuf[1]);
		System.out.println(dtmfpacket.hasMarker());
		dtmfpacket.setMarker(true);
		System.out.println(dtmfbuf[1]);
		System.out.println(dtmfpacket.hasMarker());
		dtmfpacket.setMarker(false);
		System.out.println(dtmfbuf[1]);
		System.out.println(dtmfpacket.hasMarker());
		
   }
   */

}
